/**
 * Copyright (C) 2011-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.cloudsigma.compute;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.compute.Volume;
import org.dasein.cloud.compute.VolumeState;
import org.dasein.cloud.compute.VolumeSupport;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.cloudsigma.CloudSigma;
import org.jclouds.cloudsigma.CloudSigmaAsyncClient;
import org.jclouds.cloudsigma.CloudSigmaClient;
import org.jclouds.cloudsigma.domain.ClaimType;
import org.jclouds.cloudsigma.domain.Device;
import org.jclouds.cloudsigma.domain.Drive;
import org.jclouds.cloudsigma.domain.DriveInfo;
import org.jclouds.cloudsigma.domain.DriveType;
import org.jclouds.cloudsigma.domain.ServerInfo;
import org.jclouds.rest.RestContext;

import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;

public class CSDrive implements VolumeSupport {
    static private final Logger logger = Logger.getLogger(CSDrive.class);
    
    private CloudSigma cloud;
    
    CSDrive(@Nonnull CloudSigma cloud) { this.cloud = cloud; }

    private ProviderContext getContext() throws CloudException {
        ProviderContext ctx = cloud.getContext();

        if( ctx == null ) {
            throw new CloudException("No context was established for this request");
        }
        return ctx;
    }
    
    @Override
    public void attach(@Nonnull String volumeId, @Nonnull String toServer, @Nonnull String device) throws InternalException, CloudException {
        // TODO Auto-generated method stub
        
    }

    @Override
    public @Nonnull String create(@Nullable String fromSnapshot, @Nonnegative int sizeInGb, @Nonnull String inZone) throws InternalException, CloudException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        try {
            try {
                Drive drive = (new Drive.Builder()).name("vol-" + sizeInGb).size(sizeInGb * 1024000000).build();
                DriveInfo newDrive =  ctx.getApi().createDrive(drive);
                
                if( newDrive != null ) {
                    return newDrive.getUuid();
                }
            }
            catch( Exception e ) {
                logger.error("Unable to create drive: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
            throw new CloudException("Unable to create drive without comment.");
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void detach(@Nonnull String volumeId) throws InternalException, CloudException {
        // TODO Auto-generated method stub
        
    }

    @Override
    public @Nonnull String getProviderTermForVolume(@Nonnull Locale locale) {
        return "drive";
    }

    @Override
    public @Nullable Volume getVolume(@Nonnull String volumeId) throws InternalException, CloudException {
        return toVolume(cloud.getComputeServices().getImageSupport().getDrive(volumeId));
    }

    @Override
    public @Nonnull Iterable<String> listPossibleDeviceIds(@Nonnull Platform platform) throws InternalException, CloudException {
        return Collections.emptyList(); // TODO: something better
    }

    @Override
    public @Nonnull Iterable<Volume> listVolumes() throws InternalException, CloudException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            ArrayList<Volume> volumes = new ArrayList<Volume>();
            Collection<String> drives;
            
            try {
                drives = ctx.getApi().listDrives();
            }
            catch( Exception e ) {
                logger.error("Unable to load drives: " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
            for( String driveId : drives ) {
                DriveInfo drive;
                
                try {
                    drive = ctx.getApi().getDriveInfo(driveId);
                }
                catch( Exception e ) {
                    logger.error("Unable to load drives: " + e.getMessage());
                    if( logger.isDebugEnabled() ) {
                        e.printStackTrace();
                    }
                    throw new CloudException(e);                        
                }
                Volume volume = toVolume(drive);
                    
                if( volume != null ) {
                    volumes.add(volume);
                }
            }
            return volumes;
        }
        finally {
            ctx.close();
        } 
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                ctx.getApi().listDrives();
                return true;
            }
            catch( Exception e ) {
                if( e.getMessage().contains("errors:user") || e.getMessage().contains("errors:password") ) {
                    return false;
                }
                logger.warn("Could not check subscription status in CloudSigma for " + getContext().getAccountNumber() + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        } 
    }

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }

    @Override
    public void remove(@Nonnull String volumeId) throws InternalException, CloudException {
        Volume volume = getVolume(volumeId);
        
        if( volume == null ) {
            throw new CloudException("No such volume: " + volumeId);
        }
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        try {
            try {
                ctx.getApi().destroyDrive(volumeId);
            }
            catch( Exception e ) {
                logger.error("Unable to remove drive " + volumeId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);                 
            }
        }
        finally {
            ctx.close();
        } 
    }

    private @Nullable Volume toVolume(@Nullable DriveInfo drive) throws CloudException, InternalException {
        if( drive == null ) {
            return null;
        }
        if( !drive.getClaimType().equals(ClaimType.EXCLUSIVE) ) {
            return null;
        }
        if( drive.getType().equals(DriveType.DISK) && drive.getDriveType().contains("preinstalled") ) {
            return null;
        }
        
        Volume volume = new Volume();

        volume.setProviderRegionId(getContext().getRegionId());
        volume.setProviderDataCenterId(getContext().getRegionId() + "-a");
        volume.setProviderSnapshotId(null);
        volume.setProviderVolumeId(drive.getUuid());

        Set<String> ids = drive.getClaimed();
        
        if( ids.size() < 1 ) {
            volume.setProviderVirtualMachineId(null);
            volume.setDeviceId(null);
        }
        else {
            String serverId = ids.iterator().next();
            
            volume.setProviderVirtualMachineId(serverId);
            ServerInfo server = cloud.getComputeServices().getVirtualMachineSupport().getServer(serverId);
            
            for( Map.Entry<String, Device> device : server.getDevices().entrySet() ) {
                Device d = device.getValue();
                
                if( d != null ) {
                    if( drive.getUuid().equals(d.getDriveUuid()) ) {
                        volume.setDeviceId(device.getKey());
                        break;
                    }
                }
            }
        }
        volume.setSizeInGigabytes((int)(drive.getSize()/1000000));
        volume.setName(drive.getName());
        switch( drive.getStatus() ) {
        case ACTIVE: volume.setCurrentState(VolumeState.AVAILABLE);
        case INACTIVE: case UNRECOGNIZED: volume.setCurrentState(VolumeState.DELETED); break;
        default: volume.setCurrentState(VolumeState.PENDING); break;
        }
        
        volume.setCreationTimestamp(-1L);
        return volume;
    }
}
