/**
 * Copyright (C) 2011-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.cloudsigma.compute;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.TimeUnit;

import org.apache.log4j.Logger;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.Tag;
import org.dasein.cloud.compute.Architecture;
import org.dasein.cloud.compute.MachineImage;
import org.dasein.cloud.compute.Platform;
import org.dasein.cloud.compute.VirtualMachine;
import org.dasein.cloud.compute.VirtualMachineProduct;
import org.dasein.cloud.compute.VirtualMachineSupport;
import org.dasein.cloud.compute.VmState;
import org.dasein.cloud.compute.VmStatistics;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.cloudsigma.CloudSigma;
import org.jclouds.cloudsigma.CloudSigmaAsyncClient;
import org.jclouds.cloudsigma.CloudSigmaClient;
import org.jclouds.cloudsigma.domain.Device;
import org.jclouds.cloudsigma.domain.DriveInfo;
import org.jclouds.cloudsigma.domain.DriveStatus;
import org.jclouds.cloudsigma.domain.Server;
import org.jclouds.cloudsigma.domain.ServerInfo;
import org.jclouds.cloudsigma.domain.ServerMetrics;
import org.jclouds.cloudsigma.options.CloneDriveOptions;
import org.jclouds.cloudsigma.predicates.DriveClaimed;
import org.jclouds.cloudsigma.util.Servers;
import org.jclouds.predicates.RetryablePredicate;
import org.jclouds.rest.RestContext;

import com.google.common.base.Predicates;

import javax.annotation.Nonnull;

public class CSServer implements VirtualMachineSupport {
    static private final Logger logger = Logger.getLogger(CSServer.class);
    
    private CloudSigma cloud;
    
    CSServer(CloudSigma cloud) { this.cloud = cloud; }
    
    @Override
    public void boot(String vmId) throws InternalException, CloudException {
        VirtualMachine vm = getVirtualMachine(vmId);
        
        if( vm == null ) {
            throw new CloudException("No such server: " + vmId);
        }
        
        
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                ctx.getApi().startServer(vmId);
            }
            catch( Exception e ) {
                logger.error("Could not start server " + vmId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        } 
    }

    @Override
    public VirtualMachine clone(String vmId, String intoDcId, String name, String description, boolean powerOn, String... firewallIds) throws InternalException, CloudException {
        ServerInfo server = getServer(vmId);

        if( server == null) {
            throw new CloudException("No such server: " + vmId);
        }
        VirtualMachine vm = toVirtualMachine(server);
        
        if( !vm.getCurrentState().equals(VmState.PAUSED) ) {
            throw new CloudException("You must pause a server before imaging it.");
        }
        VirtualMachineProduct product = vm.getProduct();
        String oldBootId = null;
        DriveInfo bootVolume = null;
        
        oldBootId = server.getBootDeviceIds().iterator().next();
        
        String[] parts = product.getProductId().split(":");
        int ram, cpu;
        
        try {
            ram = Integer.parseInt(parts[0]);
            cpu = Integer.parseInt(parts[1]);
        }
        catch( Throwable t ) {
            throw new CloudException("No such product: " + product.getProductId());
        }
        
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            bootVolume = ctx.getApi().getDriveInfo(oldBootId);
            bootVolume = ctx.getApi().cloneDrive(bootVolume.getUuid(), bootVolume.getUuid(), new CloneDriveOptions());
        }
        finally {
            ctx.close();
        }
        while( !bootVolume.getStatus().equals(DriveStatus.ACTIVE) ) {
            try { Thread.sleep(15000L); }
            catch( InterruptedException e ) { }
            try { bootVolume = cloud.getComputeServices().getImageSupport().getDrive(bootVolume.getUuid()); }
            catch( Throwable ignore ) { }
        }
        ctx = cloud.getCloudClient();
        try {
            boolean success =  new RetryablePredicate<DriveInfo>(Predicates.not(new DriveClaimed(ctx.getApi())), 1800, 1, TimeUnit.SECONDS).apply(bootVolume);

            if( !success ) {
                throw new CloudException("Drive from " + oldBootId + " (" + bootVolume.getUuid() + ") never became ready.");
            }
            String password = generateRandomPassword();
            
            Server s = Servers.small(name, bootVolume.getUuid(), password).mem(ram).cpu(cpu).build();
            
            vm = toVirtualMachine(ctx.getApi().createServer(s));
            if( powerOn ) {
                ctx.getApi().startServer(vm.getProviderVirtualMachineId());
            }
            return vm;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void disableAnalytics(String vmId) throws InternalException, CloudException {
        // NO-OP
    }

    @Override
    public void enableAnalytics(String vmId) throws InternalException, CloudException {
        // NO-OP
    }

    static private final Random random = new Random();
    
    static public String alphabet = "ABCEFGHJKMNPRSUVWXYZabcdefghjkmnpqrstuvwxyz#@&()=+/";
    
    public String generateRandomPassword() {
        StringBuilder password = new StringBuilder();
        int rnd = random.nextInt();
        int length = 8;
        
        if( rnd < 0 ) {
            rnd = -rnd;
        }
        length = length + (rnd%8);
        while( password.length() < length ) {
            char c;
            
            rnd = random.nextInt();
            if( rnd < 0 ) {
                rnd = -rnd;
            }
            c = (char)(rnd%255);
            if( alphabet.contains(String.valueOf(c)) ) {
                password.append(c);
            }
        }
        return password.toString();
    }
    
    @Override
    public String getConsoleOutput(String vmId) throws InternalException, CloudException {
        return "";
    }

    @Override
    public VirtualMachineProduct getProduct(String productId) throws InternalException, CloudException {
        String[] parts = productId.split(":");
        
        if( parts.length != 2 ) {
            return null;
        }
        NumberFormat fmt = NumberFormat.getInstance(Locale.getDefault());
        int ram, cpu;
        
        try {
            ram = Integer.parseInt(parts[0]);
            cpu = Integer.parseInt(parts[1]);
        }
        catch( NumberFormatException e ) {
            return null;
        }
        fmt.setMaximumFractionDigits(2);
        fmt.setMinimumIntegerDigits(1);
        VirtualMachineProduct product = new VirtualMachineProduct();
        
        product.setProductId(productId);
        product.setCpuCount(1);
        product.setDiskSizeInGb(10);
        if( ram < 1000 ) {
            product.setName(ram + "MB RAM / 1x" + fmt.format(((double)cpu)/1000.0) + "GHz CPU");
        }
        else {
            product.setName(fmt.format(((float)ram)/1024.0f) + "GB RAM / 1x" + fmt.format(((double)cpu)/1000.0) + "GHz CPU");
        }
        product.setDescription(product.getName());
        product.setRamInMb(ram);
        return product;
    }

    @Override
    public String getProviderTermForServer(Locale locale) {
        return "server";
    }

    ServerInfo getServer(String vmId) throws CloudException, InternalException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
            
        try {
            try {
                return ctx.getApi().getServerInfo(vmId);
            }
            catch( Exception e ) {
                logger.error("Could not get server info for " + vmId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }        
    }
    
    @Override
    public VirtualMachine getVirtualMachine(String vmId) throws InternalException, CloudException {
        return toVirtualMachine(getServer(vmId));
    }

    public String getVirtualMachineIdForIp(String address) throws CloudException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            for( ServerInfo server : ctx.getApi().listServerInfo() ) {
                String ip = server.getVnc().getIp();
                
                if( ip != null && ip.equals(address) ) {
                    return server.getUuid();
                }
            }
            return null;
        }
        finally {
            ctx.close();
        }
    }
    
    @Override
    public VmStatistics getVMStatistics(String vmId, long from, long to) throws InternalException, CloudException {
        ServerInfo server = getServer(vmId);
        
        if( server == null ) {
            throw new CloudException("No such server: " + vmId);
        }
        ServerMetrics metrics = server.getMetrics();
        VmStatistics stats = new VmStatistics();
        
        stats.setAverageNetworkIn(metrics.getRx());
        stats.setMaximumNetworkIn(metrics.getRx());
        stats.setMinimumNetworkIn(metrics.getRx());
        stats.setAverageNetworkOut(metrics.getTx());
        stats.setMaximumNetworkOut(metrics.getTx());
        stats.setMinimumNetworkOut(metrics.getTx());
        return stats;
    }

    @Override
    public Iterable<VmStatistics> getVMStatisticsForPeriod(String vmId, long from, long to) throws InternalException, CloudException {
        return Collections.singleton(getVMStatistics(vmId, from, to));
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                ctx.getApi().listServers();
                return true;
            }
            catch( Exception e ) {
                if( e.getMessage().contains("errors:user") || e.getMessage().contains("errors:password") ) {
                    return false;
                }
                logger.warn("Could not check subscription status in CloudSigma for " + cloud.getContext().getAccountNumber() + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }        
    }

    @Override
    public VirtualMachine launch(String fromMachineImageId, VirtualMachineProduct product, String dataCenterId, String name, String description, String withKeypairId, String inVlanId, boolean withAnalytics, boolean asSandbox, String... firewalls) throws InternalException, CloudException {
        return launch(fromMachineImageId, product, dataCenterId, name, description, withKeypairId, inVlanId, withAnalytics, asSandbox, firewalls, new Tag[0]);
    }
    
    public VirtualMachine launch(String fromMachineImageId, VirtualMachineProduct product, String dataCenterId, String name, String description, String withKeypairId, String inVlanId, boolean withAnalytics, boolean asSandbox, String[] firewalls, Tag ... tags) throws InternalException, CloudException {
        DriveInfo drive = cloud.getComputeServices().getImageSupport().getDrive(fromMachineImageId);
        
        if( drive == null ) {
            throw new CloudException("Invalid machine image ID: " + fromMachineImageId);
        }
        String[] parts = product.getProductId().split(":");
        int ram, cpu;
        
        try {
            ram = Integer.parseInt(parts[0]);
            cpu = Integer.parseInt(parts[1]);
        }
        catch( Throwable t ) {
            throw new CloudException("No such product: " + product.getProductId());
        }
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            drive = ctx.getApi().cloneDrive(drive.getUuid(), drive.getUuid(), new CloneDriveOptions());
        }
        finally {
            ctx.close();
        }
        while( !drive.getStatus().equals(DriveStatus.ACTIVE) ) {
            try { Thread.sleep(15000L); }
            catch( InterruptedException e ) { }
            try { drive = cloud.getComputeServices().getImageSupport().getDrive(drive.getUuid()); }
            catch( Throwable ignore ) { }
        }
        ctx = cloud.getCloudClient();
        try {
            boolean success =  new RetryablePredicate<DriveInfo>(Predicates.not(new DriveClaimed(ctx.getApi())), 1800, 1, TimeUnit.SECONDS).apply(drive);
            
            if( !success ) {
                throw new CloudException("Drive from " + fromMachineImageId + " (" + drive.getUuid() + ") never became ready.");
            }
            Server server = Servers.small(name, drive.getUuid(), withKeypairId).mem(ram).cpu(cpu).build();
            
            if( withKeypairId == null ) {
                withKeypairId = generateRandomPassword();
            }
            VirtualMachine vm = toVirtualMachine(ctx.getApi().createServer(server));
            ctx.getApi().startServer(vm.getProviderVirtualMachineId());
            return vm;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public Iterable<String> listFirewalls(String vmId) throws InternalException, CloudException {
        return Collections.emptyList();
    }

    static private ArrayList<VirtualMachineProduct> cachedProducts = null;
    
    @Override
    public Iterable<VirtualMachineProduct> listProducts(Architecture architecture) throws InternalException, CloudException {
        ArrayList<VirtualMachineProduct> products = cachedProducts;
        
        if( products == null ) {
            products = new ArrayList<VirtualMachineProduct>();
            
            for( int ram : new int[] { 512, 1024, 2048, 4096, 8192, 12288, 16384, 20480, 24576, 28668, 32768 } ) {
                for( int cpu : new int[] { 1000, 1200, 1500, 2000, 2500, 3000, 4000, 5000, 6000, 7000, 8000, 9000, 10000, 12000, 14000, 16000, 18000, 20000} ) {
                    products.add(getProduct(ram + ":" + cpu));
                }
            }
            cachedProducts = products;
        }
        return products;
    }

    @Override
    public Iterable<VirtualMachine> listVirtualMachines() throws InternalException, CloudException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            ArrayList<VirtualMachine> servers = new ArrayList<VirtualMachine>();
            
            for( String serverId : ctx.getApi().listServers() ) {
                ServerInfo server = ctx.getApi().getServerInfo(serverId);
                VirtualMachine vm = toVirtualMachine(server);
                    
                if( vm != null ) {
                    servers.add(vm);
                }
            }
            return servers;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }

    @Override
    public void pause(String vmId) throws InternalException, CloudException {
        VirtualMachine vm = getVirtualMachine(vmId);
        
        if( vm == null ) {
            throw new CloudException("No such server: " + vmId);
        }
        
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                ctx.getApi().stopServer(vmId);
            }
            catch( Exception e ) {
                logger.error("Could not pause server " + vmId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        } 
    }

    @Override
    public void reboot(String vmId) throws CloudException, InternalException {
        VirtualMachine vm = getVirtualMachine(vmId);
        
        if( vm == null ) {
            throw new CloudException("No such server: " + vmId);
        }
        
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                ctx.getApi().resetServer(vmId);
            }
            catch( Exception e ) {
                logger.error("Could not reboot server " + vmId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        } 
    }

    @Override
    public boolean supportsAnalytics() throws CloudException, InternalException {
        return true;
    }

    @Override
    public void terminate(String vmId) throws InternalException, CloudException {
        VirtualMachine vm = getVirtualMachine(vmId);
        
        if( vm == null ) {
            throw new CloudException("No such server: " + vmId);
        }
        
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                Server server = getServer(vmId);
                
                Map<String,Device> devices = server.getDevices();
                
                ctx.getApi().stopServer(vmId);
                try { Thread.sleep(5000L); }
                catch( InterruptedException e ) { }
                ctx.getApi().destroyServer(vmId);
                
                for( Device device : devices.values() ) {
                    if( device != null ) {
                        String driveId = device.getDriveUuid();
                        
                        if( driveId != null ) {
                            DriveInfo drive = cloud.getComputeServices().getImageSupport().getDrive(driveId);
                    
                            if( drive != null ) {
                                ctx.getApi().destroyDrive(drive.getUuid());
                            }
                        }
                    }
                }
            }
            catch( Exception e ) {
                logger.error("Could not terminate server " + vmId + ": " + e.getMessage());
                if( logger.isDebugEnabled() ) {
                    e.printStackTrace();
                }
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        } 
    }

    private VirtualMachine toVirtualMachine(ServerInfo server) throws InternalException, CloudException {
        if( server == null ) {
            return null;
        }
        VirtualMachine vm = new VirtualMachine();
        
        vm.setRootUser("root");
        vm.setRootPassword(server.getVnc().getPassword());
        vm.setProviderOwnerId(server.getUser());
        vm.setProviderRegionId(cloud.getContext().getRegionId());
        vm.setProviderDataCenterId(cloud.getContext().getRegionId() + "-a");
        vm.setProviderAssignedIpAddressId(null);

        MachineImage img = null;
        
        for( Map.Entry<String,Device> entry : server.getDevices().entrySet() ) {
            DriveInfo drive = cloud.getComputeServices().getImageSupport().getDrive(entry.getValue().getDriveUuid());
            
            if( drive != null ) {
                img = cloud.getComputeServices().getImageSupport().getMachineImage(drive.getName());
                if( img != null ) {
                    break;
                }
            }
        }
        if( img == null ) {
            vm.setArchitecture(Architecture.I64); 
            vm.setPlatform(Platform.UNKNOWN);
            vm.setProviderMachineImageId("unknown");
        }
        else {
            vm.setArchitecture(img.getArchitecture());
            vm.setPlatform(img.getPlatform());
            vm.setProviderMachineImageId(img.getProviderMachineImageId());
        }

        String ip = server.getVnc().getIp();
        boolean pub = false;
        
        if( ip != null && ip.equals("auto") ) {
            ip = null;
        }
        if( ip != null ) {
            if( !ip.startsWith("10.") && !ip.startsWith("192.168.") ) {
                if( ip.startsWith("172.") ) {
                    String[] parts = ip.split("\\.");
                    
                    if( parts.length != 4 ) {
                        pub = true;
                    }
                    else {
                        try {
                            int x = Integer.parseInt(parts[1]);
                            
                            if( x < 16 || x > 31 ) {
                                pub = true;
                            }
                        }
                        catch( NumberFormatException ignore ) {
                            // ignore
                        }
                    }
                }
                else {
                    pub = true;
                }
            }
        }
        if( pub ) {
            vm.setPrivateIpAddresses(new String[0]);            
            vm.setPublicIpAddresses( new String[] { ip });
            vm.setProviderAssignedIpAddressId(cloud.getNetworkServices().getIpAddressSupport().getIpAddressIdMatching(ip));
        }
        else {
            vm.setPrivateIpAddresses(new String[] { ip });
            vm.setPublicIpAddresses(new String[0]);
        }
        vm.setProduct(getProduct(server.getMem() + ":" + server.getCpu())); 
        vm.setImagable(false);
        vm.setLastPauseTimestamp(-1L);
        vm.setPausable(true);
        vm.setPersistent(true);
        vm.setPrivateDnsAddress(null); 
        vm.setPublicDnsAddress(null);
        vm.setRebootable(true);
        vm.setTerminationTimestamp(-1L);
        vm.setProviderVirtualMachineId(server.getUuid());
        vm.setName(server.getName());
        vm.setDescription(vm.getName());
        Date d = server.getStarted();
        
        if( d != null ) {
            vm.setLastBootTimestamp(d.getTime());
        }
        else {
            vm.setLastBootTimestamp(-1L);
        }
        vm.setCreationTimestamp(vm.getLastBootTimestamp());
        switch( server.getStatus() ) {
        case ACTIVE: vm.setCurrentState(VmState.RUNNING); break;
        case DEAD: case DUMPED: vm.setCurrentState(VmState.TERMINATED); break;
        case PAUSED: case STOPPED: vm.setCurrentState(VmState.PAUSED); break;
        default: vm.setCurrentState(VmState.PENDING); break;
        }
        vm.setClonable(vm.getCurrentState().equals(VmState.PAUSED));
        return vm;
    }
}
