/**
 * Copyright (C) 2011-2012 enStratus Networks Inc
 *
 * ====================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ====================================================================
 */

package org.dasein.cloud.jclouds.cloudsigma.network;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Locale;

import org.dasein.cloud.CloudException;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.ProviderContext;
import org.dasein.cloud.identity.ServiceAction;
import org.dasein.cloud.jclouds.cloudsigma.CloudSigma;
import org.dasein.cloud.network.AddressType;
import org.dasein.cloud.network.IpAddress;
import org.dasein.cloud.network.IpAddressSupport;
import org.dasein.cloud.network.IpForwardingRule;
import org.dasein.cloud.network.Protocol;
import org.jclouds.cloudsigma.CloudSigmaAsyncClient;
import org.jclouds.cloudsigma.CloudSigmaClient;
import org.jclouds.cloudsigma.domain.StaticIPInfo;
import org.jclouds.rest.AuthorizationException;
import org.jclouds.rest.RestContext;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

public class StaticIP implements IpAddressSupport {
    private CloudSigma cloud;
    
    StaticIP(@Nonnull CloudSigma cloud) { this.cloud = cloud; }
    
    private ProviderContext getContext() throws CloudException {
        ProviderContext ctx = cloud.getContext();
        
        if( ctx == null ) {
            throw new CloudException("No context was established for this request");
        }
        return ctx;
    }
    
    @Override
    public void assign(@Nonnull String addressId, @Nonnull String serverId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Cannot assign IP addresses via Dasein Cloud for CloudSigma");
    }

    @Override
    public @Nonnull String forward(@Nonnull String addressId, int publicPort, @Nonnull Protocol protocol, int privatePort, @Nonnull String onServerId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("IP forwarding is not supported in CloudSigma");
    }

    @Override
    public @Nullable IpAddress getIpAddress(@Nonnull String addressId) throws InternalException, CloudException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            StaticIPInfo ip = ctx.getApi().getStaticIPInfo(addressId);

            return toIpAddress(addressId, ip);
        }
        finally {
            ctx.close();
        }        
    }

    public @Nullable String getIpAddressIdMatching(@Nonnull String address) throws InternalException, CloudException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            for( String id : ctx.getApi().listStaticIPs() ) {
                StaticIPInfo ip = ctx.getApi().getStaticIPInfo(id);

                if( ip != null && ip.getAddress().equals(address) ) {
                    return id;
                }
            }
            return null;
        }
        finally {
            ctx.close();
        }        
    }
    
    @Override
    public @Nonnull String getProviderTermForIpAddress(@Nonnull Locale locale) {
        return "static IP";
    }

    @Override
    public boolean isAssigned(@Nonnull AddressType type) {
        return false;
    }

    @Override
    public boolean isForwarding() {
        return false;
    }

    @Override
    public boolean isRequestable(@Nonnull AddressType type) {
        return type.equals(AddressType.PUBLIC);
    }

    @Override
    public boolean isSubscribed() throws CloudException, InternalException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            try {
                ctx.getApi().listStaticIPInfo();
                return true;
            }
            catch( AuthorizationException e ) {
                return false;
            }
            catch( RuntimeException e ) {
                throw new CloudException(e);
            }
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public @Nonnull Iterable<IpAddress> listPrivateIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        return Collections.emptyList();
    }

    @Override
    public @Nonnull Iterable<IpAddress> listPublicIpPool(boolean unassignedOnly) throws InternalException, CloudException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            ArrayList<IpAddress> addresses = new ArrayList<IpAddress>();
            
            for( String id : ctx.getApi().listStaticIPs() ) {
                StaticIPInfo ip = ctx.getApi().getStaticIPInfo(id);
                IpAddress address = toIpAddress(id, ip);

                if( address != null ) {
                    if( !unassignedOnly || (address.getServerId() == null && address.getProviderLoadBalancerId() == null) ) {
                        addresses.add(address);
                    }
                }
            }
            return addresses;
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public @Nonnull Iterable<IpForwardingRule> listRules(@Nonnull String addressId) throws InternalException, CloudException {
        return Collections.emptyList();
    }

    @Override
    public @Nonnull String[] mapServiceAction(@Nonnull ServiceAction action) {
        return new String[0];
    }
    
    @Override
    public void releaseFromPool(@Nonnull String addressId) throws InternalException, CloudException {
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            ctx.getApi().destroyStaticIP(addressId);
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void releaseFromServer(@Nonnull String addressId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("Cannot assign IP addresses via Dasein Cloud for CloudSigma");
    }

    @Override
    public @Nonnull String request(@Nonnull AddressType typeOfAddress) throws InternalException, CloudException {
        if( typeOfAddress.equals(AddressType.PRIVATE) ) {
            throw new OperationNotSupportedException("Cannot request private IP addresses");
        }
        RestContext<CloudSigmaClient,CloudSigmaAsyncClient> ctx = cloud.getCloudClient();
        
        try {
            StaticIPInfo ip = ctx.getApi().createStaticIP();
            
            return ip.getAddress();
        }
        finally {
            ctx.close();
        }
    }

    @Override
    public void stopForward(@Nonnull String ruleId) throws InternalException, CloudException {
        throw new OperationNotSupportedException("No support for IP forwarding in CloudSigma");
    }

    private @Nullable IpAddress toIpAddress(@Nullable String id, @Nullable StaticIPInfo ip) throws CloudException {
        if( ip == null || id == null ) {
            return null;
        }
        String regionId = getContext().getRegionId();
        
        if( regionId == null ) {
            throw new CloudException("No region is set for this context");
        }
        IpAddress address = new IpAddress();
        
        address.setAddress(ip.getAddress());
        address.setAddressType(AddressType.PUBLIC);
        address.setIpAddressId(id);
        address.setProviderLoadBalancerId(null);
        address.setRegionId(regionId);
        String serverId = cloud.getComputeServices().getVirtualMachineSupport().getVirtualMachineIdForIp(ip.getAddress());
        
        if( serverId != null ) {
            address.setServerId(serverId);
        }
        return address;
    }
}
